/** <title>NSSound</title>

   <abstract>Load, manipulate and play sounds</abstract>

   Copyright (C) 2002 Free Software Foundation, Inc.
   
   Author: Enrico Sersale <enrico@imago.ro>
   Date: Jul 2002

   This file is part of the GNUstep GUI Library.
   
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; see the file COPYING.LIB.
   If not, see <http://www.gnu.org/licenses/> or write to the 
   Free Software Foundation, 51 Franklin Street, Fifth Floor, 
   Boston, MA 02110-1301, USA.
*/ 

#include <Foundation/Foundation.h>
#include <unistd.h>

#include "NSSound_test.h"
#include "GSSoundSource.h"
#include "GSSoundSink.h"

#include "SndfileSource.m"
#include "AudioOutputSink.m"

/* Private NSConditionLock conditions used for streaming */
enum
{
  SOUND_SHOULD_PLAY = 1,
  SOUND_SHOULD_PAUSE
};

#define BUFFER_SIZE 4096

/* Class variables and functions for class methods */
static NSMutableDictionary *nameDict = nil;
static NSDictionary *nsmapping = nil;

@interface NSSound (PrivateMethods)

- (void)_stream;
- (void)_finished: (NSNumber *)finishedPlaying;

@end

@implementation NSSound (PrivateMethods)

- (void)_stream
{
  NSUInteger bytesRead;
  BOOL success = YES;
  void *buffer;
  
  id<GSSoundSource> source = (id<GSSoundSource>)_private[0];
  id<GSSoundSink> sink = (id<GSSoundSink>)_private[1];
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Open sink for writing.
  [sink open];
  
  // Allocate space for buffer and start writing.
  buffer = NSZoneMalloc(NSDefaultMallocZone(), BUFFER_SIZE);
  bytesRead = [source readBytes: buffer length: BUFFER_SIZE];
  do
    {
      while ((!_shouldStop) && (bytesRead > 0) && success)
        {
          // If not SOUND_SHOULD_PLAY block thread
          if ([lock condition] != SOUND_SHOULD_PLAY)
            {
              [lock lockWhenCondition: SOUND_SHOULD_PLAY];
              [lock unlock];
              // Stop loop immediately if should stop.
              if (_shouldStop)
                {
                  break;
                }
            }
          success = [sink writeBytes: buffer length: bytesRead];
          bytesRead = [source readBytes: buffer length: BUFFER_SIZE];
        }
      
      [source setCurrentTime: 0.0];
      _shouldStop = NO;
    } while (_loops == YES); /* FIXME: The question here is,
                                 what if someone called -stop? */
  
  /* FIXME: should I send -sound:didFinishPlaying: to delegate on -stop? */
  if (!_shouldStop)
    {
      [self performSelectorOnMainThread: @selector(_finished:)
                             withObject: [NSNumber numberWithBool: success]
                          waitUntilDone: NO];
    }
  
  DESTROY(lock);
  NSZoneFree (NSDefaultMallocZone(), buffer);
  [sink close];
}

- (void)_finished: (NSNumber *)finishedPlaying
{
  [_delegate sound: self didFinishPlaying: [finishedPlaying boolValue]];
}

@end

@implementation	NSSound

+ (void) initialize
{
  if (self == [NSSound class])
    {
      NSString *path = [NSBundle pathForLibraryResource: @"nsmapping"
	                        			                 ofType: @"strings"
				                                    inDirectory: @"Sounds"];
      [self setVersion: 2];

      nameDict = [[NSMutableDictionary alloc] initWithCapacity: 10];
      
      if (path)
        {
          nsmapping = RETAIN([[NSString stringWithContentsOfFile: path]
                  propertyListFromStringsFileFormat]);
        }
      
      /* FIXME: Find and load .nssound bundles */
    }
}

- (void) dealloc
{
  // Make sure sound is stopped before deallocating.
  [self stop];
  
  TEST_RELEASE (_data);
  if ((_name != nil) && self == [nameDict objectForKey: _name])
    { 
      [nameDict removeObjectForKey: _name];
    }
  TEST_RELEASE (_name);
  TEST_RELEASE (_playbackDeviceIdentifier);
  TEST_RELEASE (_channelMapping);
  TEST_RELEASE ((id<GSSoundSource>)_private[0]);
  TEST_RELEASE ((id<GSSoundSink>)_private[1]);
  
  [super dealloc];
}

//
// Creating an NSSound 
//
- (id) initWithContentsOfFile: (NSString *)path byReference:(BOOL)byRef
{
  _onlyReference = byRef;
  ASSIGN (_name, [path lastPathComponent]);
  [nameDict setObject: self forKey: _name];
  
  {
    NSData *fileData;
    
    fileData = [NSData dataWithContentsOfMappedFile: path];
    if (!fileData)
      {
        NSLog (@"Could not get sound data from: %@", path);
        DESTROY(self);
        return nil;
      }
    
    [self initWithData: fileData];
    RELEASE(fileData);
  }
	
  return self;
}

- (id) initWithContentsOfURL: (NSURL *)url byReference:(BOOL)byRef
{
  _onlyReference = byRef;	
  return [self initWithData: [NSData dataWithContentsOfURL: url]];
}

- (id) initWithData: (NSData *)data
{
  id<GSSoundSource> source;
  id<GSSoundSink> sink;
  
  _data = data;
  RETAIN(_data);
  
  /* FIXME: need to find best GSSoundSource for this data */
  
  source = [[SndfileSource alloc] initWithData: _data];
  if (source == nil)
    {
      NSLog (@"Could not read sound data!");
      DESTROY(self);
      return nil;
    }
  _private[0] = (void *)source;
  
  sink = [[AudioOutputSink alloc] initWithEncoding: [source encoding]
                                          channels: [source channelCount]
                                        sampleRate: [source sampleRate]
                                         byteOrder: [source byteOrder]];
  if (sink == nil)
    {
      NSLog (@"Could not open sound sink!");
      DESTROY(self);
      return nil;
    }
  _private[1] = (void *)sink;
  
  return self;
}

//
// Playing and Information
//
- (BOOL) pause 
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Do nothing if sound is already paused.
  if ([lock condition] == SOUND_SHOULD_PAUSE)
    {
      return NO;
    }
  
  if ([lock tryLock] == NO)
    {
      return NO;
    }
  [lock unlockWithCondition: SOUND_SHOULD_PAUSE];
  return YES;
}

- (BOOL) play
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // If the NSCondtionLock exists it's because NSSound is either playing
  if (lock != nil)
    {
      return NO;
    }
  
  lock = [[NSConditionLock alloc] initWithCondition: SOUND_SHOULD_PAUSE];
  _private[2] = (void *)lock;
  if ([lock tryLock] != YES)
    {
      return NO;
    }
  [NSThread detachNewThreadSelector: @selector(_stream)
                           toTarget: self
                         withObject: nil];
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  
  return YES;
}

- (BOOL) resume
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  
  // Do nothing if sound is already playing.
  if ([lock condition] == SOUND_SHOULD_PLAY)
    {
      return NO;
    }
  
  if ([lock tryLock] == NO)
    {
      return NO;
    }
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  return YES;
}

- (BOOL) stop
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  if (lock == nil)
    {
      return NO;
    }
  
  if ([lock tryLock] != YES)
    {
      return NO;
    }
  _shouldStop = YES;
  // Set to SOUND_SHOULD_PLAY so that thread isn't blocked.
  [lock unlockWithCondition: SOUND_SHOULD_PLAY];
  
  sleep(1);
  /* FIXME: Do I need to wait before releasing the lock? */
  DESTROY(lock);
  _private[2] = (void *)lock;
  
  return YES;
}

- (BOOL) isPlaying
{
  NSConditionLock *lock = (NSConditionLock*)_private[2];
  if (lock == nil)
    {
      return NO;
    }
  if ([lock condition] == SOUND_SHOULD_PLAY)
    {
      return YES;
    }
  return NO;
}

- (float) volume
{
  return [(id<GSSoundSource>)_private[0] volume];
}

- (void) setVolume: (float) volume
{
  [(id<GSSoundSource>)_private[0] setVolume: volume];
}

- (NSTimeInterval) currentTime;
{
  return [(id<GSSoundSource>)_private[0] currentTime];
}

- (void) setCurrentTime: (NSTimeInterval) currentTime
{
  [(id<GSSoundSource>)_private[0] setCurrentTime: currentTime];
}

- (BOOL) loops;
{
  return NO;
}

- (void) setLoops: (BOOL) loops
{
  return;
}

- (NSTimeInterval) duration
{
  id<GSSoundSource> source = (id<GSSoundSource>)_private[0];
  return [source duration];
}

//
// Working with delegates 
//
- (id) delegate
{
  return _delegate;
}

- (void) setDelegate: (id)aDelegate
{
  _delegate = aDelegate;
}

//
// Naming Sounds 
//
+ (id) soundNamed: (NSString*)name
{
  NSString	*realName = [nsmapping objectForKey: name];
  NSSound	*sound;

  if (realName)
    {
      name = realName;
    }
	
  sound = (NSSound *)[nameDict objectForKey: name];
 
  if (sound == nil)
    {
      NSString	*extension;
      NSString	*path = nil;
      NSBundle	*main_bundle;
      NSArray	*array;
      NSString	*the_name = name;

      // FIXME: This should use [NSBundle pathForSoundResource], but this will 
      // only allow soundUnfilteredFileTypes.
      /* If there is no sound with that name, search in the main bundle */
			
      main_bundle = [NSBundle mainBundle];
      extension = [name pathExtension];
		
      if (extension != nil && [extension length] == 0)
	{
	  extension = nil;
	}

      /* Check if extension is one of the sound types */
      array = [NSSound soundUnfilteredFileTypes];
	
      if ([array indexOfObject: extension] != NSNotFound)
	{
	  /* Extension is one of the sound types
	     So remove from the name */
	  the_name = [name stringByDeletingPathExtension];
	} 
      else 
	{
	  /* Otherwise extension is not an sound type
	     So leave it alone */
	  the_name = name;
	  extension = nil;
	}

      /* First search locally */
      if (extension)
	{
	  path = [main_bundle pathForResource: the_name ofType: extension];
	} 
      else 
	{
	  id o, e;

	  e = [array objectEnumerator];
	  while ((o = [e nextObject]))
	    {
	      path = [main_bundle pathForResource: the_name ofType: o];
	      if (path != nil && [path length] != 0)
		{
		  break;
		}
	    }
	}

      /* If not found then search in system */
      if (!path)
	{
	  if (extension)
	    {
	      path = [NSBundle pathForLibraryResource: the_name
				               ofType: extension
				          inDirectory: @"Sounds"];
	    } 
	  else 
	    {
	      id o, e;

	      e = [array objectEnumerator];
	      while ((o = [e nextObject])) {
	      path = [NSBundle pathForLibraryResource: the_name
				               ofType: o
				          inDirectory: @"Sounds"];
		if (path != nil && [path length] != 0)
		  {
		    break;
		  }
	      }
	    }
	}

      if ([path length] != 0)
	{
	  sound = [[self allocWithZone: NSDefaultMallocZone()]
		    initWithContentsOfFile: path byReference: NO];

	  if (sound != nil)
	    {
	      [sound setName: name];
	      RELEASE(sound);	
	      sound->_onlyReference = YES;
	    }

	  return sound;
	}
    }  
	
  return sound;
}

+ (NSArray *) soundUnfilteredFileTypes
{
  /* FIXME */
  return nil;
}

+ (NSArray *) soundUnfilteredTypes
{
  /* FIXME */
  return nil;
}

- (NSString *) name
{
  return _name;
}

- (BOOL) setName: (NSString *)aName
{
  BOOL retained = NO;
  
  if (!aName || [nameDict objectForKey: aName])
    {
      return NO;
    }
	
  if ((_name != nil) && self == [nameDict objectForKey: _name])
    {
      /* We retain self in case removing from the dictionary releases us */
      RETAIN (self);
      retained = YES;
      [nameDict removeObjectForKey: _name];
    }
  
  ASSIGN(_name, aName);
  
  [nameDict setObject: self forKey: _name];
  if (retained)
    {
      RELEASE (self);
    }
  
  return YES;
}

- (NSString *) playbackDeviceIdentifier
{
  return [(id<GSSoundSink>)_private[1] playbackDeviceIdentifier];
}

- (void) setPlaybackDeviceIdentifier: (NSString *)playbackDeviceIdentifier
{
  return [(id<GSSoundSink>)_private[1]
           setPlaybackDeviceIdentifier: playbackDeviceIdentifier];
}

- (NSArray *) channelMapping
{
  return [(id<GSSoundSink>)_private[1] channelMapping];
}

- (void) setChannelMapping: (NSArray *)channelMapping
{
  return [(id<GSSoundSink>)_private[1] setChannelMapping: channelMapping];
}

@end
